import { FASTElementDefinition, } from "@microsoft/fast-element";
import { DI, Registration } from "../di/di";
import { ComponentPresentation } from "./component-presentation";
/**
 * @alpha
 */
export const DesignSystemRegistrationContext = DI.createInterface();
const elementTypesByTag = new Map();
const elementTagsByType = new Map();
const designSystemKey = DI.createInterface(x => x.cachedCallback(handler => {
    const element = document.body;
    const owned = element.$designSystem;
    if (owned) {
        return owned;
    }
    return new DefaultDesignSystem(element, handler);
}));
/**
 * @alpha
 */
export const DesignSystem = Object.freeze({
    tagFor(type) {
        return elementTagsByType.get(type);
    },
    responsibleFor(element) {
        const owned = element.$designSystem;
        if (owned) {
            return owned;
        }
        const container = DI.findResponsibleContainer(element);
        return container.get(designSystemKey);
    },
    getOrCreate(element = document.body) {
        const owned = element.$designSystem;
        if (owned) {
            return owned;
        }
        const container = DI.getOrCreateDOMContainer(element);
        if (!container.has(designSystemKey, false)) {
            container.register(Registration.instance(designSystemKey, new DefaultDesignSystem(element, container)));
        }
        return container.get(designSystemKey);
    },
});
class DefaultDesignSystem {
    constructor(host, container) {
        this.host = host;
        this.container = container;
        this.prefix = "fast";
        this.disambiguate = () => null;
        host.$designSystem = this;
        container.register(Registration.callback(DesignSystemRegistrationContext, () => this.context));
    }
    withPrefix(prefix) {
        this.prefix = prefix;
        return this;
    }
    withElementDisambiguation(callback) {
        this.disambiguate = callback;
        return this;
    }
    register(...registrations) {
        const container = this.container;
        const elementDefinitionEntries = [];
        const disambiguate = this.disambiguate;
        const context = (this.context = {
            elementPrefix: this.prefix,
            tryDefineElement(name, type, callback) {
                let elementName = name;
                let foundByName = elementTypesByTag.get(elementName);
                while (foundByName && elementName) {
                    elementName = disambiguate(elementName, type, foundByName);
                    if (elementName) {
                        foundByName = elementTypesByTag.get(elementName);
                    }
                }
                const willDefine = !!elementName;
                if (willDefine) {
                    if (elementTagsByType.has(type)) {
                        type = class extends type {
                        };
                    }
                    elementTypesByTag.set(elementName, type);
                    elementTagsByType.set(type, elementName);
                }
                elementDefinitionEntries.push(new ElementDefinitionEntry(container, elementName || name, type, callback, willDefine));
            },
        });
        container.register(...registrations);
        for (const entry of elementDefinitionEntries) {
            entry.callback(entry);
            if (entry.willDefine && entry.definition !== null) {
                entry.definition.define();
            }
        }
        return this;
    }
}
class ElementDefinitionEntry {
    constructor(container, name, type, callback, willDefine) {
        this.container = container;
        this.name = name;
        this.type = type;
        this.callback = callback;
        this.willDefine = willDefine;
        this.definition = null;
    }
    definePresentation(presentation) {
        ComponentPresentation.define(this.name, presentation, this.container);
    }
    defineElement(definition) {
        this.definition = new FASTElementDefinition(this.type, Object.assign(Object.assign({}, definition), { name: this.name }));
    }
    tagFor(type) {
        return DesignSystem.tagFor(type);
    }
}
